// (C) Copyright Carnegie Mellon University 2006
// All Rights Reserved.
// This code is published under the Common Public License.
//

#ifndef MyTNLP_HPP
#define MyTNLP_HPP
#include "BonTMINLP.hpp"
using namespace  Ipopt;
using namespace Bonmin;

using namespace std;
#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <math.h>
#include <ctime>

/** A C++ example for interfacing an MINLP with bonmin.
   * This class implements the following NLP :
  * \f[ 
    \begin{array}{l}
    \min - x_0 - x_1 - x_2 \\ 
    \mbox{s.t}\\
    (x_1 - \frac{1}{2})^2 + (x_2 - \frac{1}{2})^2 \leq \frac{1}{4} \\
    x_0 - x_1 \leq 0 \\
    x_0 + x_2 + x_3 \leq 2\\
    x_0 \in \{0,1\}^n \; (x_1, x_2) \in R^2 \; x_3 \in N
    \end{array}
    \f]
  */
    
    //Constants related to size of problem
    #define M 2;
    //static const int M = 2;    //Subsea manifolds
    static const int I = 6;    //Platform wells
    static const int J = 10;   //Subsea wells
    static const int J1= 6;    //Subsea wells Manifold 1
    static const int J2= 4;    //Subsea wells Manifold 2
    static const int L = 4;    //Pipelines
    static const int P = 3;    //Phases
    static const int S = 3;    //Separators   

    static const int NV = 150;  //Total number of variables
    static const int NB = 50;   //Number of binary variables
    static const int MC = 139;  //Number of constraints
    
    static const double SeA = 10;
    static const double SeB = 12;
    static const double SeC = 15;
    
    static const double A2C = 0.96; //Alternative 2 coefficient
    
    //static const int M = 2;    //Subsea manifolds
 
    static const int L1 = 2;
    static const int L2 = 2;

class MyTMINLP : public TMINLP
{
    
private:
    //Dimension of the welldata files:
    static const int gdim = 12;  //gas injection break points,
    static const int pdim = 12;  //p^W break points
    static const int ddim = 3;   //columns in the data file

    //Dimension of the pressuredrop files
    static const int dim = 30;
    static const int xdim = 30;
    static const int ydim = 30;
    static const int zdim = 30;
    
    //Separator Pressures, set in function readWells
    double PSEPA;              
    double PSEPB;
    double PSEPC;

    //Platform Well Data matrices
    double P01[gdim][pdim][ddim]; 
    double P02[gdim][pdim][ddim]; 
    double P03[gdim][pdim][ddim]; 
    double P04[gdim][pdim][ddim]; 
    double P05[gdim][pdim][ddim]; 
    double P06[gdim][pdim][ddim]; 

    //Subsea Well Data matrices
    double S01[gdim][pdim][ddim];
    double S02[gdim][pdim][ddim];
    double S03[gdim][pdim][ddim];
    double S04[gdim][pdim][ddim]; 
    double S05[gdim][pdim][ddim]; 
    double S06[gdim][pdim][ddim];
    double S07[gdim][pdim][ddim];
    double S08[gdim][pdim][ddim]; 
    double S09[gdim][pdim][ddim]; 
    double S010[gdim][pdim][ddim];

    //Fill inn WC and GOR information
    double wellspecI[I][2];
    double wellspecJ[J][2];
    void readWspecs();

    //Other
    int itt;                       //Used in simulator and printsome to count how many times the simulator is called
    double minmax_w[I+J][3][2];    //Array containing min and max values for all wells minmax_w[well][qGL/pW/oil][min/max]
    
public:
    static int a;
    static int b;
    
    //Functions to read Well data
    void readWells(); //called from main to read all the well data
    void readWellData(const string &filen, double data[gdim][pdim][ddim]);//called from readWells() -> reads and saves one well
    void minmax_wells(int, double data[gdim][pdim][ddim]);//Saves the min/max values into the array minmax_w
    void set_arrptr();

    //Functions to read Pressure Drop data
    double gas[dim];
    double oil[dim];
    double wat[dim];
    double dP1[dim][dim][dim][4];
    double dP2[dim][dim][dim][4];
    double dP3[dim][dim][dim][4];
    double dP4[dim][dim][dim][4];
    void readDeltaPData(const string &filen, double data[dim][dim][dim][4]); //called form readPipes() -> reads and saves one pipeline
    void readPipes(); //called from main to read all the pipeline data

    //Making an array to hold pointers to the Pipeline Tables
    typedef double (*dPtrArray)[dim][dim][4];
    dPtrArray dParrptr[L];
    void set_dParrPtr();

    typedef double (*PtrArray)[pdim][ddim];
    //Making an array to hold pointers to the Platform Well Tables
    PtrArray parrptr[I];

    //Making an array to hold pointers to the Subsea Well Tables
    PtrArray sarrptr[J];

    //Functions to write output files
    void printsome(int, int, int, double, double, double, double, double val[12]); //for  printe ut og sjekke datainnlesing, kan slettes etterhvert
    void set_itt(){itt=0;};

    //Simulators
    double * ratesim(int, double, double, double (*)[pdim][ddim]);
    double * PressureDrop(int, double, double, double, double (*)[dim][dim][4]);

    //General Functions
    int round(double);
    int min_val(int, double a[]);//returnerer index til minste
    int max_val(int, double a[]);//returnerer index til strste
    int closest_zero(int, double a[]);//returnerer index til positivt tall nrmest null

    double bounds[NV][2];
    double start_pnt[NV];
    double solution[NV];
    double obj;
    double s;
    const double* const_sol;
    
    
  /// Default constructor.
  MyTMINLP():
printSol_(false){}
  
  /// virtual destructor.
  virtual ~MyTMINLP(){}

  
	/** Copy constructor.*/   
  MyTMINLP(const MyTMINLP &other):
printSol_(other.printSol_){}
  /** Assignment operator. no data = nothing to assign*/
  //MyTMINLP& operator=(const MyTMINLP&) {}

  
  /** \name Overloaded functions specific to a TMINLP.*/
  //@{
  /** Pass the type of the variables (INTEGER, BINARY, CONTINUOUS) to the optimizer.
     \param n size of var_types (has to be equal to the number of variables in the problem)
  \param var_types types of the variables (has to be filled by function).
  */
  virtual bool get_variables_types(Index n, VariableType* var_types);
 
  /** Pass info about linear and nonlinear variables.*/
  virtual bool get_variables_linearity(Index n, Ipopt::TNLP::LinearityType* var_types);

  /** Pass the type of the constraints (LINEAR, NON_LINEAR) to the optimizer.
  \param m size of const_types (has to be equal to the number of constraints in the problem)
  \param const_types types of the constraints (has to be filled by function).
  */
  virtual bool get_constraints_linearity(Index m, Ipopt::TNLP::LinearityType* const_types);
//@}  
    
  /** \name Overloaded functions defining a TNLP.
     * This group of function implement the various elements needed to define and solve a TNLP.
     * They are the same as those in a standard Ipopt NLP problem*/
  //@{
  /** Method to pass the main dimensions of the problem to Ipopt.
        \param n number of variables in problem.
        \param m number of constraints.
        \param nnz_jac_g number of nonzeroes in Jacobian of constraints system.
        \param nnz_h_lag number of nonzeroes in Hessian of the Lagrangean.
        \param index_style indicate wether arrays are numbered from 0 (C-style) or
        from 1 (Fortran).
        \return true in case of success.*/
  virtual bool get_nlp_info(Index& n, Index&m, Index& nnz_jac_g,
                            Index& nnz_h_lag, TNLP::IndexStyleEnum& index_style);
  
  /** Method to pass the bounds on variables and constraints to Ipopt. 
       \param n size of x_l and x_u (has to be equal to the number of variables in the problem)
       \param x_l lower bounds on variables (function should fill it).
       \param x_u upper bounds on the variables (function should fill it).
       \param m size of g_l and g_u (has to be equal to the number of constraints in the problem).
       \param g_l lower bounds of the constraints (function should fill it).
       \param g_u upper bounds of the constraints (function should fill it).
  \return true in case of success.*/
  virtual bool get_bounds_info(Index n, Number* x_l, Number* x_u,
                               Index m, Number* g_l, Number* g_u);
  
  /** Method to to pass the starting point for optimization to Ipopt.
    \param init_x do we initialize primals?
    \param x pass starting primal points (function should fill it if init_x is 1).
    \param m size of lambda (has to be equal to the number of constraints in the problem).
    \param init_lambda do we initialize duals of constraints? 
    \param lambda lower bounds of the constraints (function should fill it).
    \return true in case of success.*/
  virtual bool get_starting_point(Index n, bool init_x, Number* x,
                                  bool init_z, Number* z_L, Number* z_U,
                                  Index m, bool init_lambda,
                                  Number* lambda);
  
  /** Method which compute the value of the objective function at point x.
    \param n size of array x (has to be the number of variables in the problem).
    \param x point where to evaluate.
    \param new_x Is this the first time we evaluate functions at this point? 
    (in the present context we don't care).
    \param obj_value value of objective in x (has to be computed by the function).
    \return true in case of success.*/
  virtual bool eval_f(Index n, const Number* x, bool new_x, Number& obj_value);

  /** Method which compute the gradient of the objective at a point x.
    \param n size of array x (has to be the number of variables in the problem).
    \param x point where to evaluate.
    \param new_x Is this the first time we evaluate functions at this point? 
    (in the present context we don't care).
    \param grad_f gradient of objective taken in x (function has to fill it).
    \return true in case of success.*/
  virtual bool eval_grad_f(Index n, const Number* x, bool new_x, Number* grad_f);

  /** Method which compute the value of the functions defining the constraints at a point
    x.
    \param n size of array x (has to be the number of variables in the problem).
    \param x point where to evaluate.
    \param new_x Is this the first time we evaluate functions at this point? 
    (in the present context we don't care).
    \param m size of array g (has to be equal to the number of constraints in the problem)
    \param grad_f values of the constraints (function has to fill it).
    \return true in case of success.*/
  virtual bool eval_g(Index n, const Number* x, bool new_x, Index m, Number* g);

  /** Method to compute the Jacobian of the functions defining the constraints.
    If the parameter values==NULL fill the arrays iCol and jRow which store the position of
    the non-zero element of the Jacobian.
    If the paramenter values!=NULL fill values with the non-zero elements of the Jacobian.
    \param n size of array x (has to be the number of variables in the problem).
    \param x point where to evaluate.
    \param new_x Is this the first time we evaluate functions at this point? 
    (in the present context we don't care).
    \param m size of array g (has to be equal to the number of constraints in the problem)
    \param grad_f values of the constraints (function has to fill it).
    \return true in case of success.*/
  virtual bool eval_jac_g(Index n, const Number* x, bool new_x,
                          Index m, Index nele_jac, Index* iRow, Index *jCol,
                          Number* values);
  
  /** Method to compute the Jacobian of the functions defining the constraints.
    If the parameter values==NULL fill the arrays iCol and jRow which store the position of
    the non-zero element of the Jacobian.
    If the paramenter values!=NULL fill values with the non-zero elements of the Jacobian.
    \param n size of array x (has to be the number of variables in the problem).
    \param x point where to evaluate.
    \param new_x Is this the first time we evaluate functions at this point? 
    (in the present context we don't care).
    \param m size of array g (has to be equal to the number of constraints in the problem)
    \param grad_f values of the constraints (function has to fill it).
    \return true in case of success.*/
  virtual bool eval_h(Index n, const Number* x, bool new_x,
                      Number obj_factor, Index m, const Number* lambda,
                      bool new_lambda, Index nele_hess, Index* iRow,
                      Index* jCol, Number* values);

  
  /** Method called by Ipopt at the end of optimization.*/  
  virtual void finalize_solution(TMINLP::SolverReturn status,
                                 Index n, const Number* x, Number obj_value);
  
  //@}

  virtual const SosInfo * sosConstraints() const{return NULL;}
  virtual const BranchingInfo* branchingInfo() const{return NULL;}
  
  
  void printSolutionAtEndOfAlgorithm(){
    printSol_ = true;}
  
private:
   bool printSol_;
};

#endif
